@@ -6,7 +6,6 @@ from typing import TYPE_CHECKING, Any, Callable
 from ._pycrdt import Array as _Array
 from ._pycrdt import ArrayEvent as _ArrayEvent
 from .base import BaseDoc, BaseEvent, BaseType, base_types, event_types
-from .transaction import ReadTransaction
 
 if TYPE_CHECKING:
     from .doc import Doc
@@ -169,8 +168,15 @@ class Array(BaseType):
 
     def to_py(self) -> list | None:
         if self._integrated is None:
-            return self._prelim
-        return list(self)
+            py = self._prelim
+            if py is None:
+                return None
+        else:
+            py = list(self)
+        for idx, val in enumerate(py):
+            if isinstance(val, BaseType):
+                py[idx] = val.to_py()
+        return py
 
     def observe(self, callback: Callable[[Any], None]) -> str:
         _callback = partial(observe_callback, callback, self.doc)
@@ -190,19 +196,15 @@ class Array(BaseType):
 
 def observe_callback(callback: Callable[[Any], None], doc: Doc, event: Any):
     _event = event_types[type(event)](event, doc)
-    doc._txn = ReadTransaction(doc=doc, _txn=event.transaction)
-    callback(_event)
-    doc._txn = None
+    with doc._read_transaction(event.transaction):
+        callback(_event)
 
 
 def observe_deep_callback(callback: Callable[[Any], None], doc: Doc, events: list[Any]):
     for idx, event in enumerate(events):
         events[idx] = event_types[type(event)](event, doc)
-        if idx == 0:
-            _event = event
-    doc._txn = ReadTransaction(doc=doc, _txn=_event.transaction)
-    callback(events)
-    doc._txn = None
+    with doc._read_transaction(event.transaction):
+        callback(events)
 
 
 class ArrayEvent(BaseEvent):
