@@ -6,7 +6,6 @@ from typing import TYPE_CHECKING, Any, Callable
 from ._pycrdt import Map as _Map
 from ._pycrdt import MapEvent as _MapEvent
 from .base import BaseDoc, BaseEvent, BaseType, base_types, event_types
-from .transaction import ReadTransaction
 
 if TYPE_CHECKING:
     from .doc import Doc
@@ -63,8 +62,15 @@ class Map(BaseType):
 
     def to_py(self) -> dict | None:
         if self._integrated is None:
-            return self._prelim
-        return dict(self)
+            py = self._prelim
+            if py is None:
+                return None
+        else:
+            py = dict(self)
+        for key, val in py.items():
+            if isinstance(val, BaseType):
+                py[key] = val.to_py()
+        return py
 
     def __delitem__(self, key: str) -> None:
         with self.doc.transaction() as txn:
@@ -150,19 +156,15 @@ class Map(BaseType):
 
 def observe_callback(callback: Callable[[Any], None], doc: Doc, event: Any):
     _event = event_types[type(event)](event, doc)
-    doc._txn = ReadTransaction(doc=doc, _txn=event.transaction)
-    callback(_event)
-    doc._txn = None
+    with doc._read_transaction(event.transaction):
+        callback(_event)
 
 
 def observe_deep_callback(callback: Callable[[Any], None], doc: Doc, events: list[Any]):
     for idx, event in enumerate(events):
         events[idx] = event_types[type(event)](event, doc)
-        if idx == 0:
-            _event = event
-    doc._txn = ReadTransaction(doc=doc, _txn=_event.transaction)
-    callback(events)
-    doc._txn = None
+    with doc._read_transaction(event.transaction):
+        callback(events)
 
 
 class MapEvent(BaseEvent):
