@@ -14,9 +14,9 @@ class Transaction:
     _txn: _Transaction | None
     _nb: int
 
-    def __init__(self, doc: Doc, _txn: _Transaction | None = None) -> None:
+    def __init__(self, doc: Doc) -> None:
         self._doc = doc
-        self._txn = _txn
+        self._txn = None
         self._nb = 0
 
     def __enter__(self) -> Transaction:
@@ -35,7 +35,7 @@ class Transaction:
         self._nb -= 1
         # only drop the transaction when exiting root context manager
         # since nested transactions reuse the root transaction
-        if self._nb == 0 and not isinstance(self, ReadTransaction):
+        if self._nb == 0:
             # dropping the transaction will commit, no need to do it
             # self._txn.commit()
             assert self._txn is not None
@@ -45,4 +45,19 @@ class Transaction:
 
 
 class ReadTransaction(Transaction):
-    pass
+    def __init__(self, doc: Doc, _txn: _Transaction) -> None:
+        self._doc = doc
+        self._txn = _txn
+
+    def __enter__(self) -> Transaction:
+        self._doc._txn = self
+        return self
+
+    def __exit__(
+        self,
+        exc_type: type[BaseException] | None,
+        exc_val: BaseException | None,
+        exc_tb: TracebackType | None,
+    ) -> None:
+        self._txn = None
+        self._doc._txn = None
